.campbell-mode-shape {
    display: flex;
    flex-wrap: wrap;
    width: 100vw;
    height: 100vh;
    justify-content: center;
}

.campbell-mode-shape > * {
    width: 100%;
    height: 50%;
    min-height: 768px;
    padding: 5px;
}

.campbell-mode-shape > *:first-child {
    min-height: 468px;
    width: 98%;
}

@media (min-width: 1200px) {
    .campbell-mode-shape > * {
        height: 100%;
        width: 40%;
    }
    
    .campbell-mode-shape > *:first-child {
        width: 60%;
    }
}
